"""Main package API."""

from typing import Any
from typing import List
from typing import Optional
from typing import Union

from numpy.typing import NDArray

from .errors import DimensionMismatchError
from .errors import TooManyDimensionsError
from .utils import _parse_lines
from .utils import use_clipboard


@use_clipboard
def to_matrix(
    arr: NDArray[Any],
    style: str = "b",
    num_format: Optional[str] = None,
    scientific_notation: bool = False,
    to_clp: bool = False,
) -> str:
    """Convert a numpy.NDArray to LaTeX matrix.

    Args:
        arr: the array to be converted
        style: a style formatter string, either "b" for "bmatrix" or "p" for "pmatrix"
        num_format: a number formatter string, e.g. ".2f"
        scientific_notation: a flag to determine whether 1 x 10^3 should be used,
            otherwise e-notation is used (1e3)
        to_clp: copy the output to the system clipboard

    Returns:
        the string representation of the array

    Raises:
        TooManyDimensionsError: when the supplied array has more than 2 dimensions
    """
    if len(arr.shape) > 2:
        raise TooManyDimensionsError

    environment = f"{style}matrix"

    lines = _parse_lines(arr, num_format, scientific_notation)

    rv = [f"\\begin{{{environment}}}"]
    rv += [line.strip() + r" \\" for line in lines]
    rv += [f"\\end{{{environment}}}"]

    return "\n".join(rv)


@use_clipboard
def to_tabular(
    arr: NDArray[Any],
    num_format: Optional[str] = None,
    scientific_notation: bool = False,
    col_align: Union[List[str], str] = "c",
    cols: Optional[List[str]] = None,
    to_clp: bool = False,
) -> str:
    """Convert a numpy.NDArray to LaTeX tabular environment.

    Args:
        arr: the array to be converted
        num_format: a number formatter string, e.g. ".2f"
        scientific_notation: a flag to determine whether 1 x 10^3 should be used,
            otherwise e-notation is used (1e3)
        col_align: set the alignment of the columns, usually "c", "r" or "l". If a
            single character is provided then it will be broadcast to all columns. If a list
            is provided then each item will be assigned to each column, list size and
            number of columns must match
        cols: an optional list of column names, otherwise generic names will be assigned
        to_clp: copy the output to the system clipboard

    Returns:
        the string representation of the array

    Raises:
        TooManyDimensionsError: when the supplied array has more than 2 dimensions
        DimensionMismatchError: when there are mismatched column identifiers and
            dimensions
    """
    n_dims = len(arr.shape)

    if n_dims == 0:
        n_cols = 1
    elif n_dims == 1:
        n_cols = arr.shape[0]
    elif n_dims == 2:
        n_cols = arr.shape[1]
    else:
        raise TooManyDimensionsError

    if isinstance(col_align, list) and len(col_align) != n_cols:
        raise DimensionMismatchError(
            "Number of `col_align` items doesn't match number of columns "
            + f"({len(col_align)} against {n_cols})"
        )

    if isinstance(col_align, str):
        col_align = [col_align for _ in range(n_cols)]

    if cols and len(cols) != n_cols:
        raise DimensionMismatchError(
            "Number of `cols` items doesn't match number of columns "
            + f"({len(cols)} against {n_cols})"
        )

    if not cols:
        cols = [f"Col {i + 1}" for i in range(n_cols)]

    lines = _parse_lines(arr, num_format, scientific_notation)

    rv = [f"\\begin{{tabular}}{{{' '.join(col_align)}}}"]
    rv += [r"\toprule"]
    rv += [" & ".join(cols) + r" \\"]
    rv += [r"\midrule"]
    rv += [line.strip() + r" \\" for line in lines]
    rv += [r"\endrule"]
    rv += [r"\end{tabular}"]

    return "\n".join(rv)



"""Utils module."""
import re
from functools import wraps
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import numpy as np
import pyperclip
from numpy.typing import NDArray
from typing_extensions import ParamSpec


P = ParamSpec("P")
T = TypeVar("T")


def use_clipboard(func: Callable[P, T]) -> Callable[P, str]:
    """Augument decorated functions argument to copy the output to the clipboard."""

    @wraps(func)
    def wrapper_func(*args: P.args, **kwargs: P.kwargs) -> str:
        """Wrapped function."""
        out = func(*args, **kwargs)

        if kwargs.get("to_clp"):
            pyperclip.copy(out)
            print("ArrayTeX: copied to clipboard")

        return str(out)

    return wrapper_func


def _parse_lines(
    arr: NDArray[Any],
    num_format: Optional[str] = None,
    scientific_notation: bool = False,
) -> List[str]:
    formatter = {}
    if num_format:

        def num_formatter(x: Union[np.int64, np.float64, np.float32]) -> str:
            return f"%{num_format}" % x

        formatter.update({"float_kind": num_formatter, "int_kind": num_formatter})

    lines = (
        np.array2string(
            arr,
            max_line_width=np.inf,  # type: ignore  # noqa
            formatter=formatter,  # type: ignore  # noqa
            separator=" & ",
        )
        .replace("[", "")
        .replace("]", "")
        .replace(" &\n", "\n")
    )

    if num_format and "e" in num_format:
        pattern = r"e(-?\d+)"
        replace = r"\\mathrm{e}{\g<1>}"
        if scientific_notation:
            replace = r" \\times 10^{\g<1>}"
        lines = re.sub(pattern, replace, lines)

    return lines.splitlines()



"""Module to hold custom errors."""


class TooManyDimensionsError(Exception):
    """Raised when an array of more than 2 dimensions is supplied."""


class DimensionMismatchError(Exception):
    """Raised when dimensionality issues occur."""
